#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

const int nReps = 100;

//Macro for checking cuda errors following a cuda launch or api call
#define cudaCheckError() {                                          \
 cudaError_t e=cudaGetLastError();                                 \
 if(e!=cudaSuccess) {                                              \
   printf("Cuda failure %s:%d: '%s'\n",__FILE__,__LINE__,cudaGetErrorString(e));           \
   exit(0); \
 }                                                                 \
}

void time_kernel(char* label, void (*fptr)(int, int, float*, float*),  int rows, int cols, float * in, float * out);
void time_kernel_cuda(char* label, void (*fptr)(int, int, float*, float*),  int rows, int cols, float * in, float * out);

void referenceTranspose(int rows, int cols, float * in, float * out)
{
  #pragma omp parallel for
  for(int i=0; i<rows; i++)
  {
   #pragma omp parallel for
   for(int j=0; j<cols; j++)
    {
      out[i*rows + j] = in[j*cols + i];
    }
  }
}


__global__ void gpuTranspose1_kernel(int rows, int cols, float * in, float * out)
{
  int i; int j;

  i = blockIdx.x * blockDim.x + threadIdx.x;
  for(j=0; j<cols; j++)
  {
    out[i*rows + j] = in[j*cols + i];
  }
}

void gpuTranspose1(int rows, int cols, float * in, float * out) {
  gpuTranspose1_kernel<<<rows/256, 256>>>(rows,cols,in,out);
  cudaCheckError();
}


int main(int argc, char** argv)
{
  static int M = 1024;
  static int N = 1024;
  float * A = NULL;
  float * Atrans = NULL;
  A = (float*)malloc(M*N*sizeof(float));
  Atrans = (float*)malloc(M*N*sizeof(float));

  printf("OpenMP Processors: %d\n",omp_get_max_threads());

  // Initialize A to random values and zero Atrans
  for(int i=0; i<M*N; i++)
  {
    A[i] = rand() / (float)RAND_MAX;
  }
  memset(Atrans, 0, M*N*sizeof(float));

  time_kernel((char*)"CPU+OMP", referenceTranspose, M, N, A, Atrans);
  time_kernel_cuda((char*)"CUDA", gpuTranspose1, M, N, A, Atrans);

  free(A);
  free(Atrans);
  return 0;
}

void time_kernel(char* label, void (*fptr)(int, int, float*, float*),  int rows, int cols, float * in, float * out) {
  double start, end;

  start = omp_get_wtime();
  for(int i=0; i<nReps; i++)
  {
    fptr(rows, cols, in, out);
  }
  end = omp_get_wtime();
  
  // Check the answer: return time only if answer is correct. This section part is not timed.
  float * validation = (float*)malloc(rows*cols*sizeof(float));
  referenceTranspose(rows, cols, in, validation);
  for(long i=0; i<rows*cols; i++)
  {
    if(out[i] != validation[i])
    {
      printf("Error: transpose result is incorrect at linear index %ld\n", i);
      return;
    }
  }
  free(validation);

  double kTime = (end - start)/((double) nReps);
  double kBandwidth = 2.0 * (double)(rows*cols*sizeof(float)) / (1000*1000*1000*kTime);
  printf("%s: Kernel bandwidth: %f gb/sec\n", label, kBandwidth);

}

void time_kernel_cuda(char* label, void (*fptr)(int, int, float*, float*),  int rows, int cols, float * in, float * out) {
  double start, end;

  /********************************************
   * Allocate device memory
   *******************************************/
  float * devA;
  float * devAtrans;
  cudaMalloc((void**)&devA, rows*cols*sizeof(float)); 
  cudaMalloc((void**)&devAtrans, rows*cols*sizeof(float)); 
  cudaCheckError();
  /********************************************
   * Copy data to the device
   *******************************************/
  cudaMemcpy(devA, in, rows*cols*sizeof(float), cudaMemcpyHostToDevice); 
  cudaCheckError();

  /********************************************
   * Ensure that device is empty of work 
   * before we start timing.
   *******************************************/
  cudaDeviceSynchronize(); 
  cudaCheckError(); 

  start = omp_get_wtime();
  for(int i=0; i<nReps; i++)
  {
    /*****************************************
     * Pass in device pointers
     ****************************************/
    fptr(rows, cols, devA, devAtrans); 
  }

  /********************************************
   * Ensure that device is finished  
   * before we stop timing.
   *******************************************/
  cudaDeviceSynchronize(); 
  cudaCheckError();

  end = omp_get_wtime();
  
  /********************************************
   * Copy results back to the host
   *******************************************/
  cudaMemcpy(out, devAtrans, rows*cols*sizeof(float), cudaMemcpyDeviceToHost); 
  cudaCheckError();

  // Check the answer: return time only if answer is correct. This section part is not timed.
  float * validation = (float*)malloc(rows*cols*sizeof(float));
  referenceTranspose(rows, cols, in, validation);
  for(long i=0; i<rows*cols; i++)
  {
    if(out[i] != validation[i])
    {
      printf("Error: transpose result is incorrect at linear index %ld\n", i);
      return;
    }
  }
  free(validation);
  /********************************************
   * Free cuda memory
   *******************************************/
  cudaFree(devA);
  cudaFree(devAtrans);
  cudaCheckError();

  double kTime = (end - start)/((double) nReps);
  double kBandwidth = 2.0 * (double)(rows*cols*sizeof(float)) / (1000*1000*1000*kTime);
  printf("%s: Kernel bandwidth: %f gb/sec\n", label, kBandwidth);
}
